#ifndef MATRIX_H_
#define MATRIX_H_

#include "Vector.h"

#include <cmath>
#include <cstring>
#include <string>
#include <sstream>

namespace arda 
   {
   namespace Math
      {
      //////////////////////////////////////////////////////////////////////////
      // Maxtrices
      //
      // Supported operations on matrices
      //
      // []
      //     If you have a Matrix named m you can access it's member elements as
      //     m[0], m[1], m[2], etc.  The indexing is COLUMN MAJOR.  So for a 2x2
      //     matrix, m[0] and m[1] are the elements of the first column, not the
      //     first row.
      // == != + += - -=
      //     Defined for operations on two matrices of the same type.
      // * *= / /=
      //     Defined for scalar multiplication/division with int, float, and double.
      //     * is defined as a standalone template operator for the scalar * Matrix
      //     form.
      // * *=
      //     Also defined for Matrix * Matrix
      //
      // assign()          Assign the value of one matrix to another.
      // getcol()          Returns the vector that is column n of the matrix.
      // getstring()       Returns a printable string representation of the matrix.
      // setcol()          Set the vector that is column n of the matrix.
      //                   Overloaded so that you can set from a vector or from
      //                   specific values.
      // setidentity()     Sets a matrix to the identity matrix.
      //
      //
      //
      // The following are functions, not class methods.  This is because I think
      // that the notation det(m) and transpose(m), for example, is more natural
      // to write and read than m.det() and m.transpose().  Using functional
      // notation just more naturally mimics the standard mathematical notation.
      //
      // det()             Calculates the determinant of a matrix.
      // transpose()       Returns the transpose of a matrix.
      
      //////////////////////////////////////////////////////////////////////////
      template <typename T> class Matrix22
	 {
	 public:
	 T m[4];

	 // Constructors
	 Matrix22() {}
	 explicit Matrix22(T a) { m[0] = m[1] = m[2] = m[3] = a; }
	 Matrix22(T a0, T a1, T a2, T a3) 
	    { 
	    m[0] = a0; m[2] = a2;
	    m[1] = a1; m[3] = a3; 
	    } 

	 // Array indexing
	 // Remember: column major order is used.
	 inline T& operator[](unsigned int i) 
	    { assert (i<4); return m[i]; }
	 inline T operator[](unsigned int i) const
	    { assert (i<4); return m[i]; }

	 // Assignment
	 inline Matrix22<T>& operator=(const Matrix22<T>& m2)
	    { 
	    m[0] = m2[0]; m[2] = m2[2]; 
	    m[1] = m2[1]; m[3] = m2[3]; 
	    return *this;
	    }
	 inline Matrix22<T>& assign(T a0, T a1, T a2, T a3)
	    { 
	    m[0] = a0; m[2] = a2;
	    m[1] = a1; m[3] = a3; 
	    return *this;
	    }

	 // Comparison
	 inline bool operator==(Matrix22<T>& m2) const
	    {
	    return  
		  m[0] == m2[0] && m[2] == m2[2] &&
		  m[1] == m2[1] && m[3] == m2[3]; 
	    }
	 inline bool operator!=(Matrix22<T>& m2) const
	    { return ! (*this == m2); } 

	 // Matrix addition
	 inline Matrix22<T>& operator+=(const Matrix22<T>& m2)
	    {
	    m[0] += m2[0]; m[2] += m2[2]; 
	    m[1] += m2[1]; m[3] += m2[3];
	    return *this;
	    }
	 inline Matrix22<T> operator+(const Matrix22<T>& m2) const
	    { return Matrix22<T>(*this) += m2; }

	 // Matrix subtraction
	 inline Matrix22<T>& operator-=(const Matrix22<T>& m2)
	    {
	    m[0] -= m2[0]; m[2] -= m2[2]; 
	    m[1] -= m2[1]; m[3] -= m2[3];
	    return *this;
	    }
	 inline Matrix22<T> operator-(const Matrix22<T>& m2) const
	    { return Matrix22<T>(*this) -= m2; }
      
	 // Scalar multiplication
	 inline Matrix22<T>& operator*=(const int a)
	    {
	    m[0] *= a; m[2] *= a;
	    m[1] *= a; m[3] *= a;
	    return *this;
	    }
	 inline Matrix22<T>& operator*=(const float a)
	    {
	    m[0] *= a; m[2] *= a;
	    m[1] *= a; m[3] *= a;
	    return *this;
	    }
	 inline Matrix22<T>& operator*=(const double a)
	    {
	    m[0] *= a; m[2] *= a;
	    m[1] *= a; m[3] *= a;
	    return *this;
	    }

	 // Scalar division
	 inline Matrix22<T>& operator/=(const int a)
	    {
	    assert(a!=0);
	    m[0] /= a; m[2] /= a;
	    m[1] /= a; m[3] /= a;
	    return *this;
	    }
	 inline Matrix22<T>& operator/=(const float a)
	    {
	    assert(a!=0);
	    m[0] /= a; m[2] /= a;
	    m[1] /= a; m[3] /= a;
	    return *this;
	    }
	 inline Matrix22<T>& operator/=(const double a)
	    {
	    assert(a!=0);
	    m[0] /= a; m[2] /= a;
	    m[1] /= a; m[3] /= a;
	    return *this;
	    }
	 inline Matrix22<T> operator/(const int a)
	    { return Matrix22<T>(*this) /= a; }
	 inline Matrix22<T> operator/(const float a)
	    { return Matrix22<T>(*this) /= a; }
	 inline Matrix22<T> operator/(const double a)
	    { return Matrix22<T>(*this) /= a; }

	 // Matrix multiplication
	 // Not sure if this should be inlined at all.  Sure the compiler will
	 // probably ignore the inline request here, but maybe it won't and maybe
	 // that would be bad.  Needs real testing.
	 inline Matrix22<T>& operator*=(Matrix22<T> m2)
	    {
	    const int size=2;
	    Matrix22<T> mres(0);
	    int i, j, k;
	    for (i=0; i<size; ++i)
	       for (j=0; j<size; ++j)
		  for (k=0; k<size; ++k)
		     mres[size*i+j] += (*this)[size*k+j] * m2[size*i+k];
	    *this = mres;
	    return *this;
	    }
	 inline Matrix22<T> operator*(Matrix22<T> m2) const
	    { return Matrix22<T>(*this) *= m2; }


	 inline Vector2<T> getcol(const unsigned int i) const
	    {
	    assert(i<2);
	    const int size=2;
	    return Vector2<T>(m[i*size], m[i*size+1]);
	    }

	 inline Vector2<T> getrow(const unsigned int i) const
	    {
	    assert(i<2);
	    const int size=2;
	    return Vector2<T>(m[i], m[i+size]);
	    }
	 
	 inline Matrix22<T>& setcol(const unsigned int i, const Vector2<T>& v)
	    {
	    assert(i<2);
	    const int size=2;
	    m[i*size] = v[0];
	    m[i*size+1] = v[1];
	    return *this;
	    }

	 inline Matrix22<T>& setcol(const unsigned int i, const T a, const T b)
	    {
	    assert(i<2);
	    const int size=2;
	    m[i*size] = a;
	    m[i*size+1] = b;
	    return *this;
	    }

	 inline Matrix22<T>& setrow(const unsigned int i, const Vector2<T>& v)
	    {
	    assert(i<2);
	    const int size=2;
	    m[i] = v[0];
	    m[i+size] = v[1];
	    return *this;
	    }

	 inline Matrix22<T>& setrow(const unsigned int i, const T a, const T b)
	    {
	    assert(i<2);
	    const int size=2;
	    m[i] = a;
	    m[i+size] = b;
	    return *this;
	    }

	 std::string & getstring(std::string & s) const;

	 inline Matrix22<T>& setidentity()
	    {
	    memset(m, 0, sizeof(m));
	    m[0] = m[3] = 1;
	    return *this;
	    }

	 };

      // Scalar multiplication continued
      template <typename T> inline Matrix22<T> operator*(const Matrix22<T>& m, const int a)
	 { return Matrix22<T>(m) *= a; }
      template <typename T> inline Matrix22<T> operator*(const int a, const Matrix22<T>& m)
	 { return Matrix22<T>(m) *= a; }
      template <typename T> inline Matrix22<T> operator*(const Matrix22<T>& m, const float a)
	 { return Matrix22<T>(m) *= a; }
      template <typename T> inline Matrix22<T> operator*(const float a, const Matrix22<T>& m)
	 { return Matrix22<T>(m) *= a; }
      template <typename T> inline Matrix22<T> operator*(const Matrix22<T>& m, const double a)
	 { return Matrix22<T>(m) *= a; }
      template <typename T> inline Matrix22<T> operator*(const double a, const Matrix22<T>& m)
	 { return Matrix22<T>(m) *= a; }




      //////////////////////////////////////////////////////////////////////////
      typedef Matrix22<int> Matrix22i;
      typedef Matrix22<float> Matrix22f;
      typedef Matrix22<double> Matrix22d;

#ifdef NOTHING   
      typedef Matrix33<int> Matrix33i;
      typedef Matrix33<float> Matrix33f;
      typedef Matrix33<double> Matrix33d;
      
      typedef Matrix44<int> Matrix44i;
      typedef Matrix44<float> Matrix44f;
      typedef Matrix44<double> Matrix44d;
      
#endif



      //////////////////////////////////////////////////////////////////////////
      // Matrix functions

      template <typename T> inline double det(const Matrix22<T> m)
	 { return (double) (m[0] * m[3]) - (double) (m[1] * m[2]); }

      template <typename T> inline Matrix22<T> transpose(const Matrix22<T>& m)
	 {
	 const int size=2;
	 Matrix22<T> mres(0);
	 int i, j;
	 for (i=0; i<size; ++i)
	    for (j=0; j<size; j++)
	       mres[size*i+j] = m[size*j+i];
	 return mres;
	 }

      } // namespace Math
   } // namespace arda


////////////////////////////////////////////////////////////////////////////////
// Matrix methods

template <typename T> std::string & arda::Math::Matrix22<T>::getstring(std::string & s) const
   {
   s.clear();
   std::stringstream ss;
   ss << "[ "
	 "[ " << m[0] << ", " << m[1] << " ], "
	 "[ " << m[2] << ", " << m[3] << " ] ]";
   s = ss.str();
   return s;   
   }












#ifdef NOTHING
   //////////////////////////////////////////////////////////////////////////
   // det*(Matrix m)
   //	Computer the determinant of a matrix.
   //	Returns the determinant.

   inline double det3(Matrix3i m)
      { return m[0] * (  m[4] * m[8]  
			 - m[5] * m[7])  
	    -  m[3] * (  m[1] * m[8]
			 - m[2] * m[7]) 
	    +  m[6] * (  m[1] * m[5] 
			 - m[2] * m[4]); }

   inline double det4(Matrix4i m)
      { return m[0] * (m[5] * (m[10] * m[15] - 
			       m[11] * m[14]) - 
		       m[9] * (m[6] * m[15] - 
			       m[7] * m[14]) +
		       m[13] * (m[6] * m[11] - 
				m[7] * m[10])) -
	    m[4] * (m[1] * (m[10] * m[15] - 
			    m[11] * m[14]) - 
		    m[9] * (m[2] * m[15] - 
			    m[3] * m[14]) +
		    m[13] * (m[2] * m[11] - 
			     m[3] * m[10])) +
	    m[8] * (m[1] * (m[6] * m[15] - 
			    m[7] * m[14]) - 
		    m[5] * (m[2] * m[15] - 
			    m[3] * m[14]) +
		    m[13] * (m[2] * m[7] - 
			     m[3] * m[6])) +
	    m[12] * (m[1] * (m[6] * m[11] - 
			     m[7] * m[10]) - 
		     m[5] * (m[2] * m[11] - 
			     m[3] * m[10]) +
		     m[9] * (m[2] * m[7] - 
			     m[3] * m[6])); }

////////////////////////////////////////////////////////////////////////////////
int * arda::Matrix::multvec3mv(Matrix3i m, Vector3i v, Vector3i vres)
   {
   vres[0] = m[0]*v[0] + m[3]*v[1] + m[6]*v[2];
   vres[1] = m[1]*v[0] + m[4]*v[1] + m[7]*v[2];
   vres[2] = m[2]*v[0] + m[5]*v[1] + m[8]*v[2];
   return vres;
   }


////////////////////////////////////////////////////////////////////////////////
int * arda::Matrix::multvec4mv(Matrix4i m, Vector4i v, Vector4i vres)
   {
   vres[0] = m[0]*v[0] + m[4]*v[1] + m[8] *v[2] + m[12]*v[3];
   vres[1] = m[1]*v[0] + m[5]*v[1] + m[9]*v[2]  + m[13]*v[3];
   vres[2] = m[2]*v[0] + m[6]*v[1] + m[10]*v[2] + m[14]*v[3];
   vres[3] = m[3]*v[0] + m[7]*v[1] + m[11]*v[2] + m[15]*v[3];
   return vres;
   }

////////////////////////////////////////////////////////////////////////////////
int * arda::Matrix::multvec3vm(Vector3i v, Matrix3i m, Vector3i vres)
   {
   vres[0] = v[0]*m[0] + v[1]*m[1] + v[2]*m[2];
   vres[1] = v[0]*m[3] + v[1]*m[4] + v[2]*m[5];
   vres[2] = v[0]*m[6] + v[1]*m[7] + v[2]*m[8];
   return vres;   
   }

////////////////////////////////////////////////////////////////////////////////
float * arda::Matrix::multvec4vm(Vector4f v, Matrix4f m, Vector4f vres)
   {
   vres[0] = v[0]*m[0]  + v[1]*m[1]  + v[2]*m[2]  + v[3]*m[3];
   vres[1] = v[0]*m[4]  + v[1]*m[5]  + v[2]*m[6]  + v[3]*m[7];
   vres[2] = v[0]*m[8]  + v[1]*m[9]  + v[2]*m[10] + v[3]*m[11];
   vres[3] = v[0]*m[12] + v[1]*m[13] + v[2]*m[14] + v[3]*m[15];
   return vres;   
   }
#endif




#endif // MATRIX_H_
